"""
Model classes used by the Workspaces system.
"""

import enum
import json
from pathlib import Path
from typing import Dict, List, Tuple, Any
import chevron

import sqlalchemy as sa
from pycapo import CapoConfig
from sqlalchemy import create_engine
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from workspaces.workflow_interfaces import (
    WorkflowIF,
    WorkflowRequestIF,
    WorkflowRequestFileIF,
)

from .capability_interfaces import (
    CapabilityIF,
    CapabilityRequestIF,
    CapabilityExecutionIF,
    CapabilitySequenceIF,
    ParameterIF,
    CapabilityVersionIF,
)
from .helpers import (
    CapabilityStepType,
    CapabilityStep,
    CapabilitySequence,
    CapabilityEventType,
)
from .product_interfaces import FutureProductIF


class AbstractFile:
    """
    Abstract file is exactly that, an abstract concept of what a file is, to be
    returned from various non-filesystem places.
    """

    def __init__(self, filename: str, content: bytes):
        self.filename, self.content = filename, content

    def write_to(self, directory: Path):
        (directory / self.filename).write_bytes(self.content)

    def __json__(self, request):
        return {"filename": self.filename}

    @classmethod
    def from_path(cls, path: Path) -> "AbstractFile":
        return cls(path.name, path.read_bytes())


class AbstractRequest:
    """
    Abstract File is an abstract concept that is used to create requests
    """

    name: str
    argument: json

    def __init__(self, name: str, argument: json):
        self.name, self.argument = name, argument

    def __json__(self, request):
        return {"workflow_name": self.name, "argument": self.argument}


class WorkflowEventType(enum.Enum):
    """
    Kinds of events that a workflow can send. These four are derived from
    the many kinds that HTCondor can send.
    """

    SUBMITTED = 0
    EXECUTING = 1
    TERMINATED = 5
    OTHER = -1


class WorkflowEvent:
    """
    An event from the underlying workflow execution system, which is likely
    to be HTCondor.
    """

    def __init__(
        self,
        workflow_request_id: int,
        job_name: str,
        condor_job_id: str,
        event_type: WorkflowEventType,
        timestamp: str,
        log: str,
        retval: int = None,
    ):
        # NOTE: when you modify the properties here, also update the schema
        # in the .json package.
        self.workflow_request_id = workflow_request_id
        self.job_name = job_name
        self.condor_job_id = str(condor_job_id)
        self.type = event_type
        self.timestamp = timestamp
        self.log = log
        self.retval = retval

    def json(self) -> str:
        d = {
            "workflow_request_id": self.workflow_request_id,
            "job_name": self.job_name,
            "type": str(self.type),
            "timestamp": self.timestamp,
            "return_value": self.retval,
            "log": self.log,
        }
        return json.dumps(d)

    def __str__(self):
        return f"WorkflowEvent with data {self.json()}"

    def __repr__(self):
        return f"<WorkflowEvent {self.__dict__}>"

    def __eq__(self, other):
        return (
            self.job_name == other.job_name
            and self.condor_job_id == other.condor_job_id
            and self.type == other.type
            and self.timestamp == other.timestamp
            and self.log == other.log
            and self.retval == other.retval
        )


class CapabilityEvent:
    """
    An event sent from the underlying workflow execution system.
    """

    def __init__(
        self, event_type: CapabilityEventType, request_id: int, execution_id: int
    ):
        self.event_type = event_type
        self.execution_id = execution_id

    def json(self) -> str:
        return json.dumps(self.__dict__)

    def __str__(self):
        return f"<CapabilityEvent with data{self.json()}>"

    def __repr__(self):
        return f"<CapabilityEvent {self.__dict__}>"

    def __eq__(self, other):
        return self.json() == other.json()


Base = declarative_base()


class Capability(Base, CapabilityIF):
    """
    A capability, which is a particular workflow setup, intended to accept
    a certain kind of product and some parameters and produce another product
    """

    __tablename__ = "capabilities"
    id = sa.Column("capability_id", sa.Integer, primary_key=True)
    name = sa.Column("capability_name", sa.String)
    steps = sa.Column("capability_steps", sa.String)
    max_jobs = sa.Column("max_jobs", sa.Integer)
    requests = relationship("CapabilityRequest", back_populates="capability")

    @classmethod
    def from_file(cls, filename: str):
        """
        Method that creates an instance of a Capability from a given file

        :param filename:
        :return: Capability instance
        """
        self = cls()
        self.id, self.name, self.max_jobs, self.steps = cls.parse_capability_file(
            filename
        )
        return self

    @classmethod
    def from_json(cls, json_data: str):
        """
        Method that creates an instance of a Capability from given JSON

        :param json_data:
        :return: Capability instance
        """
        self = cls()
        json_dict = json.loads(json_data)
        self.id = json_dict.get("id", None)
        self.name = json_dict["name"]
        self.max_jobs = json_dict["max_jobs"]
        self.steps = json_dict["steps"]
        return self

    @staticmethod
    def parse_capability_file(
        filename: str,
    ) -> Tuple[int, str, int, CapabilitySequenceIF]:
        """
        Parse a file containing metadata for a capability

        :param filename:
        :return: Tuple containing capability's ID, name, max concurrent jobs, and sequence of steps
        """
        with open(filename, "r") as f:
            cap_id, name, max_jobs = f.readline().split(" ")
            steps = []

            for line in f.readlines():
                step_type, step_value = line.strip().split(" ")
                steps.append(
                    CapabilityStep(
                        CapabilityStepType.from_string(step_type), step_value
                    )
                )

            return int(cap_id), name, int(max_jobs), CapabilitySequence(steps)

    def create_request(
        self, parameters: "ParameterIF", future_products: FutureProductIF
    ):
        """
        Create a new request for this capability with specific options

        FIXME: Future products needs a string representation
        """
        request = CapabilityRequest(capability=self, parameters=str(parameters))
        self.requests.append(request)
        return request

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "max_jobs": self.max_jobs,
            "steps": str(self.steps),
        }

    def __str__(self):
        return (
            f"Capability object with ID {self.id}, name {self.name} and max simultaneous jobs of {self.max_jobs}"
            f"\nSequence: {self.steps}"
        )


class CapabilityRequest(Base, CapabilityRequestIF):
    """
    A capability request, which couples a capability to a product, representing
    the expectation of a new product given a set of parameters
    """

    __tablename__ = "capability_requests"
    id = sa.Column("capability_request_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_id = sa.Column(
        "capability_id", sa.Integer, sa.ForeignKey("capabilities.capability_id")
    )
    parameters = sa.Column("parameters", sa.String)
    # FIXME: This needs to be changed to properly keep track of product locators.
    # future_products = sa.Column('future_products', sa.String)
    versions = relationship("CapabilityVersion", back_populates="request")
    capability = relationship(Capability, back_populates="requests")

    def update_status(self, status: str):
        # TODO: create field in table
        self.status = status

    def __str__(self):
        return f"CapabilityRequest object: {self.__dict__}"

    def __json__(self, request) -> dict:
        return dict(
            id=self.id,
            capability_id=self.capability_id,
            state=self.state,
            parameters=self.parameters,
        )


class CapabilityVersion(Base, CapabilityVersionIF):
    """
    Schema representation of a capability version
    """

    __tablename__ = "capability_versions"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
        primary_key=True,
    )
    version_number = sa.Column("version_number", sa.Integer, primary_key=True)
    parameters = sa.Column("parameters", sa.String)
    request = relationship(CapabilityRequest, back_populates="versions")
    executions = relationship("CapabilityExecution", back_populates="version")

    @property
    def capability(self):
        return self.request.capability


class CapabilityExecution(Base, CapabilityExecutionIF):
    """
    Schema representation of a capability request's execution record
    """

    __tablename__ = "capability_executions"

    id = sa.Column("execution_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
    )
    version_number = sa.Column("capability_version_number", sa.Integer)
    current_step = sa.Column("current_step", sa.Integer)
    steps = sa.Column("steps", sa.String)
    version = relationship(CapabilityVersion, back_populates="executions")
    current_workflow_request_id = sa.Column(
        "current_workflow_request_id",
        sa.Integer,
        sa.ForeignKey("workflow_requests.workflow_request_id"),
    )
    current_workflow_request = relationship("WorkflowRequest")

    __table_args__ = (
        sa.ForeignKeyConstraint(
            [capability_request_id, version_number],
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
    )

    def __json__(self, request: Any) -> dict:
        return dict(
            id=self.id,
            state=self.state,
            capability_request_id=self.capability_request_id,
            version_number=self.version_number,
            current_step=self.current_step,
            current_workflow_request_id=self.current_workflow_request_id,
        )

    @property
    def capability(self):
        return self.version.capability

    @property
    def capability_request(self):
        return self.version.request


class Workflow(Base, WorkflowIF):
    """
    A Workflow is a suite of tasks defined by templates that must be
    executed together with some user-submitted files.
    """

    __tablename__ = "workflows"
    workflow_name = sa.Column("workflow_name", sa.String, primary_key=True)
    templates = relationship("WorkflowTemplate", backref="workflows", lazy='dynamic')
    requests = relationship("WorkflowRequest", backref="workflows")

    def __json__(self, request):
        return {"workflow_name": self.workflow_name, "templates": self.templates}

    def __repr__(self):
        return f"<Workflow workflow_name={self.workflow_name}>"

    def render_templates(
        self, argument: Dict, files: List[Path] = None
    ) -> List[AbstractFile]:
        """
        Render the templates associated with this workflow

        :param argument: the workflow argument JSON
        :param files:    the files to be processed
        :return:         a list of rendered template files
        """
        rendered_files = []

        for f in files:
            rendered_files.append(AbstractFile.from_path(f))

        for template in self.templates:
            rendered_files.append(template.render(argument))

        return rendered_files


class WorkflowTemplate(Base):
    """
    Workflow Templates are Mustache-formatted files associated with a given Workflow
    """

    __tablename__ = "workflow_templates"
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)
    workflow_name = sa.Column(
        "workflow_name",
        sa.String,
        sa.ForeignKey("workflows.workflow_name"),
        primary_key=True,
    )

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Workflow Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """

        # TODO: if template has more than one tag, figure out how pass argument parameter
        #  in as JSON and replace 'data' section accordingly
        args = {
            'template': self.content.decode('ascii'),
            'data': {
                'arguments': ''.join(map(str, argument))
            }
        }
        contents = chevron.render(**args)
        return AbstractFile(self.filename, contents.encode('ascii'))

    @property
    def template_file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @template_file.setter
    def template_file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __json__(self, request):
        # Defer to the actual template file's contents for JSON conversion
        return self.template_file.__json__(request)

    def __repr__(self):
        return f"<WorkflowTemplate filename={self.filename} for workflow={self.workflow_name}>"


class WorkflowRequest(Base, WorkflowRequestIF):
    """
    Workflow Requests invoke a particular workflow with a particular argument and some user-supplied files.
    """

    __tablename__ = "workflow_requests"
    workflow_request_id = sa.Column(
        "workflow_request_id", sa.Integer, primary_key=True, autoincrement=True
    )
    workflow_name = sa.Column(
        "workflow_name", sa.String, sa.ForeignKey("workflows.workflow_name")
    )
    argument = sa.Column("argument", sa.JSON)
    state = sa.Column("state", sa.String)
    results_dir = sa.Column("results_dir", sa.String)
    files = relationship("WorkflowRequestFile", backref="request")

    @property
    def request(self) -> AbstractRequest:
        return AbstractRequest(self.workflow_name, self.argument)

    @request.setter
    def create_request(self, request: AbstractRequest):
        self.workflow_name, self.argument = request.name, request.argument

    def update_status(self, status: str):
        self.state = status

    # TODO: create following fields in table
    def set_start_time(self, time: str):
        self.start_time = time

    def set_end_time(self, time: str):
        self.end_time = time

    def __json__(self, request) -> dict:
        return dict(
            workflow_request_id=self.workflow_request_id,
            workflow_name=self.workflow_name,
            argument=self.argument,
            state=self.state,
            results_dir=self.results_dir,
            files=self.files,
        )

    def __getitem__(self, item):
        return self.files

    def __repr__(self):
        return f"<WorkflowRequest workflow_request_id= {self.workflow_request_id}>"


class WorkflowRequestFile(Base, WorkflowRequestFileIF):
    """
    A Workflow Request File is a file supplied by the user and attached to the request they have submitted.
    """

    __tablename__ = "workflow_request_files"
    workflow_request_id = sa.Column(
        "workflow_request_id",
        sa.Integer,
        sa.ForeignKey("workflow_requests.workflow_request_id"),
        primary_key=True,
    )
    filename = sa.Column("filename", sa.String, primary_key=True)
    content = sa.Column("content", sa.LargeBinary, nullable=False)

    @property
    def file(self) -> AbstractFile:
        return AbstractFile(self.filename, self.content)

    @file.setter
    def file(self, file: AbstractFile):
        self.filename, self.content = file.filename, file.content

    def __repr__(self):
        return f"<WorkflowRequestFile filename={self.filename}>"

    def __json__(self, request):
        return {"filename": self.filename}

class NotificationTemplate(Base):
    """
    Notification Templates are Mustache-formatted files used for sending emails with the notification system.
    """

    __tablename__ = "notification_templates"
    name = sa.Column("name", sa.String, primary_key=True)
    content = sa.Column("template", sa.String, nullable=False)

    def render(self, argument: Dict) -> AbstractFile:
        """
        Render a given Mustache-formatted Workflow Template with the specified arguments.

        :param argument:    argument(s) needed for template render
        :return:            an AbstractFile rendered from mustache template
        """

        # TODO: if template has more than one tag, figure out how pass argument parameter
        #  in as JSON and replace 'data' section accordingly
        args = {
            'template': self.content.decode('ascii'),
            'data': {
                'arguments': ''.join(map(str, argument))
            }
        }
        contents = chevron.render(**args)
        return contents



def get_engine():
    """
    Generate the SQL Alchemy engine for us, using Capo

    :return: SQLAlchemy engine from database URL
    """

    capo = CapoConfig().settings("metadataDatabase")
    url = capo.jdbcUrl.replace("jdbc:", "").replace(
        "://", f"://{capo.jdbcUsername}:{capo.jdbcPassword}@"
    )
    return create_engine(url)


def get_session_factory(engine):
    """
    Generate the SQL Alchemy session factory for us, using the supplied engine

    :param engine:
    :return: SQLAlchemy engine
    """

    factory = sessionmaker(expire_on_commit=False)
    factory.configure(bind=engine)
    return factory
